"""Implements functionality for Snapshot inputs."""

from disco.distribution.distribution_inputs import DistributionInputs


class PyDssInputs(DistributionInputs):
    """Implements functionality for PyDSS inputs."""
