
from disco.distribution.deployment_parameters import DeploymentParameters


class UpgradeParameters(DeploymentParameters):
    """Represents deployment parameters for upgrade simulation/analysis

    Parameters
    ----------
    DeploymentParameters : object
        Deployment Job
    """

